/****************************************************************************/
/*                    Canon PowerShot I/O library 2.5f (fixed by AlexG)     */
/****************************************************************************/
/* Created: 25.9.2004 [BORLAND C 4.52]                                      */
/* Last modified: 31.7.2005 [BORLAND C 4.52]                                */
/* Copyright (C) 2004-2005 by RayeR, Lexeich, Gambistics, Nick T Tate       */
/* Contact: http://www.volny.cz/rayer                                       */
/* Compiler: Borland C++ 4.52, disable Word Alignment, use 80186 instr. set */
/****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <dir.h>

//**************** PSIOlib user configurable variables-customize for your camera
//#define CAMERA_A70 // CAMERA_A40, CAMERA_IXUSV - define camera type for proper keycodes, define it in your program before include psiolib.c
#define PS_FONT_W 22                   // define font width; 22 for A70
#define PS_FONT_H 24                   // define font height; 24 for A70 (doubled => 1 line height is 48 pix.)
#ifndef CAMERA_S70
#define PS_SCREEN_XRES 704             // define screen x-resolution; 704 for A70, 720 for S400
#else
#define PS_SCREEN_XRES 720             // define screen x-resolution;
#endif


#define PS_SCREEN_YRES 481             // define screen y-resolution; 481 for A70, 540 for S400
                                       // aspect ratio for A70: X=1.20, Y=1.00

//**************** PSIOlib internal definitions and types *******************
#define PS_USED_API_FUNCS 16           // define how many API functions we have, usually 16
typedef unsigned char Byte;            // size 8 bit, 0-255
typedef unsigned int  Word;            // size 16 bit, 0-65535
typedef unsigned long DWord;           // size 32 bit, 0-4294967295
#define FAR far                        // keyword for far 32 bit RM pointers
#define FARMEMCPY _fmemcpy             // memcpy with far pointers, string.h
#define FARPASCAL far pascal           // Pascal-style function keyword
typedef void FAR * FARPTR;             // generic far pointer type
FARPTR ps_api_functions[PS_USED_API_FUNCS]; // pointers to API functions, need to be initialized via ps_init()
Word ps_where_x=0; Word ps_where_y=0;  // keeps current coords for ps_printf

//**************** PSIOlib structures ***************************************
typedef struct                         // text message structure
{                                      // x-position [0-703] (but LCD is 320x240)
  Word x,y;                            // y-position [0-480] means 32x10 chars
  char FAR *strptr;                    // RM pointer to 0-terminated string
  Word drawmode;                       // drawing mode (0=transparent,2=filled by 2,3=corrupted,5=bothfilled)
  Word txtcolor;                       // text color (14-white, depends on color palette)
  Word bkcolor;                        // background color, usually 0
} PS_TEXT_MSG;

typedef struct                         // system info structure
{
  DWord unknown1;                      // ? (A70: 00 00 25 01)
  DWord unknown2;                      // ? (A70: A9 C4 BC 0C)
  Byte fwver[4];                       // firmware version (A70: 00 00 00 02) - little endian
  DWord unknown3;                      // ? (A70: 2C 00 93 D0)
  Byte unknown4;                       // ? (A70: 00)
  char idstring[31];                   // full camera name
} PS_SYS_INFO;

typedef struct                         // keys (buttons) state structure
{
  Word event;                          // usually 4, change to 3 when POWER
  Word unknown1;                       // button pressed or switched to SHOOT mode
  Byte unknown2;
  DWord keymap;                        // see defined constants below
} PS_KEY_STATE;

//**************** some error codes constants *******************************
#define PS_ERR_CODE_MASK 0xFFF         // bit mask for returned status/error codes
#define PS_ERR_INACTIVITY_TIMEOUT      0x02
#define PS_ERR_DISK_INACCESSIBLE       0x30
#define PS_ERR_DISK_FAILURE            0x31
#define PS_ERR_TIMEOUT                 0x32
#define PS_ERR_NAMING_FAILURE          0x33
#define PS_ERR_NULL_PTR                0x34
#define PS_ERR_FILE_NOT_FOUND          0x40
#define PS_ERR_DOS_ERR                 0x50
#define PS_ERR_ACCESS_DENIED           0x51
#define PS_ERR_INVALID_FILE_HANDLE     0x52
#define PS_ERR_TOO_MANY_ARGS           0x53
#define PS_ERR_FILE_EXISTS             0x55
#define PS_ERR_TOO_MANY_OPEN_FILES     0x56
#define PS_ERR_FILE_NOT_FOUND_DOS      0x57
#define PS_ERR_DISK_FULL               0x5A
#define PS_ERR_OUT_OF_MEM_DOS          0x5B
#define PS_ERR_BATT_DOOR_OPEN          0x403
#define PS_ERR_BATT_DEAD               0x404
#define PS_ERR_CF_DOOR_OPEN            0x405

//**************** language IDs constants ***********************************
#define PS_LANG_ENGLISH                0x00
#define PS_LANG_GERMAN                 0x01
#define PS_LANG_FRENCH                 0x02
#define PS_LANG_DUTCH                  0x03
#define PS_LANG_DANISH                 0x04
#define PS_LANG_FINNISH                0x05 // Czech replacement
#define PS_LANG_ITALIAN                0x06
#define PS_LANG_NORWEGIAN              0x07
#define PS_LANG_SWEDISH                0x08
#define PS_LANG_SPANISH                0x09

//**************** video modes constants ************************************
#define PS_VIDEOMODE_LCD               0x00
#define PS_VIDEOMODE_NTSC              0x01 // LCD is off
#define PS_VIDEOMODE_PAL               0x02 // LCD is off

//**************** keys (buttons) constants (for A40,A6x,A7x,A8x,A9x customize yourself)
#define PS_KEY_SHUTTER_HALF            0x00000001 // SHUTTER button top/right half-pressed
#define PS_KEY_SHUTTER_FULL            0x00000002 // SHUTTER button top/right full-pressed
#define PS_KEY_CF_DOOR_CLOSED          0x00000004 // CF door on the right side
#define PS_KEY_SET                     0x00000008 // SET button under LCD
#define PS_KEY_MENU                    0x00000010 // MENU button under LCD
#define PS_KEY_DISP                    0x00000040 // DISP button under LCD
#define PS_KEY_SHOOT_MODE              0x00000100 // camera was switched to SHOOT mode
#define PS_KEY_FUNC                    0x00000200 // FUNC button under LCD


#ifdef CAMERA_S70
  #define PS_KEY_RIGHT                 0x00000400 // RIGHT cross-key button beside LCD
  #define PS_KEY_LEFT                  0x00000800 // LEFT cross-key button beside LCD
  #define PS_KEY_UP                    0x00020000 // UP AKA Flash cross-key button
  #define PS_KEY_DOWN                  0x00040000 // DOWN AKA Macro cross-key button
  #define PS_KEY_VIEW                  0x00000200 // camera was switched to SHOOT mode
#endif





#ifdef CAMERA_A40
  #define PS_KEY_RIGHT                 0x00040000 // RIGHT button below LCD
#endif
#ifdef CAMERA_A70
  #define PS_KEY_RIGHT                 0x00000400 // RIGHT cross-key button beside LCD
#endif
#ifdef CAMERA_IXUSV
  #define PS_KEY_RIGHT                 0x00040000 // RIGHT button below LCD
#endif
#ifdef CAMERA_A40
  #define PS_KEY_LEFT                  0x00020000 // LEFT button below LCD
#endif
#ifdef CAMERA_A70
  #define PS_KEY_LEFT                  0x00000800 // LEFT cross-key button beside LCD
#endif
#ifdef CAMERA_IXUSV
  #define PS_KEY_LEFT                  0x00020000 // LEFT button below LCD
#endif
#define PS_KEY_ZOOM_OUT                0x00001000 // ZOOM OUT on top/right side
#define PS_KEY_ZOOM_IN                 0x00002000 // ZOOM IN on top/right side
#define PS_KEY_BATT_DOOR_CLOSED        0x00004000 // BATTERY door on bottom side
#define PS_KEY_VIDEO_CONNECTED         0x00008000 // VIDEO cable is attached
#define PS_KEY_USB_CONNECTED           0x00010000 // USB cable is attached
#ifdef CAMERA_A40
  #define PS_KEY_UP                    0x00001000 // UP remapped to ZOOM- (then cannot use ZOOM-)
#endif
#ifdef CAMERA_A70
  #define PS_KEY_UP                    0x00020000 // UP AKA Flash cross-key button
#endif
#ifdef CAMERA_IXUSV
  #define PS_KEY_UP                    0x00040008 // UP as SET+LEFT key combo (but then SET cannot be used)
#endif
#ifdef CAMERA_A40
  #define PS_KEY_DOWN                  0x00002000 // UP remapped to ZOOM+ (then cannot use ZOOM+)
#endif
#ifdef CAMERA_A70
  #define PS_KEY_DOWN                  0x00040000 // DOWN AKA Macro cross-key button
#endif
#ifdef CAMERA_IXUSV
  #define PS_KEY_DOWN                  0x00020008 // DOWN as SET+RIGHT key combo (but then SET cannot be used)
#endif
#define PS_KEY_POWER                   0x40000000 // POWER button on top/middle
#define PS_KEY_ANY                     0x40063F5B // Any button exlude doors and cables

//**************** notes for ps_sound constants (nearest frequencies [Hz]) **
#define PS_NOTE_MINF                   61         // max. possible frequency
#define PS_NOTE_MAXF                   15625      // max. possible frequency
#define PS_NOTE_C1                     264        // N=58
#define PS_NOTE_D1                     297        // N=51
#define PS_NOTE_E1                     330        // N=46
#define PS_NOTE_F1                     352        // N=43
#define PS_NOTE_G1                     396        // N=38
#define PS_NOTE_A1                     440        // N=34
#define PS_NOTE_A1_                    466        // _ mean #, N=32
#define PS_NOTE_H1                     494        // N=30
#define PS_NOTE_C2                     520        // 523, N=29
#define PS_NOTE_C2_                    554        // N=27
#define PS_NOTE_D2                     578        // 587, N=26
#define PS_NOTE_D2_                    622        // N=24
#define PS_NOTE_E2                     651        // 659, N=23
#define PS_NOTE_F2                     698        // N=21
#define PS_NOTE_F2_                    740        // N=20
#define PS_NOTE_G2                     781        // 784, N=19
#define PS_NOTE_G2_                    822        // 831, N=18
#define PS_NOTE_A2                     868        // 880, N=17
#define PS_NOTE_A2_                    919        // 932, N=16
#define PS_NOTE_H2                     976        // 988, N=15
#define PS_NOTE_C3                     1041       // 1047, N=14
#define PS_NOTE_C3_                    1109       // N=13
#define PS_NOTE_D3                     1175       // N=12
#define PS_NOTE_D3_                    1245       // N=11
#define PS_NOTE_E3                     1302       // 1319, N=11
#define PS_NOTE_F3                     1397       // N=10
#define PS_NOTE_F3_                    1480       // N=9
#define PS_NOTE_G3                     1562       // 1568, N=9
#define PS_NOTE_G3_                    1661       // N=8
#define PS_NOTE_A3                     1736       // 1760, N=8
#define PS_NOTE_A3_                    1865       // N=7
#define PS_NOTE_H3                     1953       // 1976, N=7


//**************** mem_avail return maximum allocatable block size **********
long mem_avail(void)                   // I found that largest free memory
{                                      // you can allocate in model Tiny and Small
  long i;
  void *ptr;
  for (i=1; i<1024; i++)               // step by 1kB
    {
    ptr=malloc(i*1024);                // allocate a block of memory
    if (ptr==NULL)                     // if failed
      return((i-1)*1024);              // return last valid size
    else
      free(ptr);                       // free memory block
    }
  return(0);
}

//**************** subfunctions of API function 1 - system ******************
typedef Word (FARPASCAL PS_KEY_HANDLER)(PS_KEY_STATE FAR *keys, DWord FAR *args);

typedef struct
{
  Byte FAR *(FARPASCAL *subfunc_00)(void); // 00h (get system info)
  Byte unknown1[0x04];                 // 04-07h
  Word (FARPASCAL *subfunc_08)(PS_KEY_HANDLER FAR *handler, DWord FAR *args); // 08h (process key handler)
  Byte unknown2[0x08];                 // 0C-13h
  Word (FARPASCAL *subfunc_14)(void);  // 14h (get status)
  Byte unknown3[0x34];                 // 18-4Bh
  void (FARPASCAL *subfunc_4c)(Byte lang_id); // 4Ch (set lang id)
  Byte (FARPASCAL *subfunc_50)(void);  // 50h (get lang id)
  Byte unknown4[0x68];                 // 54-BBh
  void (FARPASCAL *subfunc_bc)(FARPTR t); // BCh (get timer - uses QWord value!)
  Byte unknown5[0x08];                 // C0-C7h
  Word (FARPASCAL *subfunc_c8)(void);  // C8h (get PID)
} PS_FUNCS_SYSTEM;
PS_FUNCS_SYSTEM FAR *ps_funcs_system_ptr;

//**************** get system status, if some error occured return error code
Word ps_get_status(void) {             // also prevent camera auto powerdown after 16s of run time
  return((ps_funcs_system_ptr->subfunc_14())&PS_ERR_CODE_MASK);
}                                      // see definitions above for returned error codes

//**************** get camera product id (stored in upid.ini) ***************
Word ps_get_pid(void) {return(ps_funcs_system_ptr->subfunc_c8());}    // 3073h for A70

//**************** set camera language id ***********************************
void ps_set_lang_id(Byte lang_id) {ps_funcs_system_ptr->subfunc_4c(lang_id);} // the change take effect after restart

//**************** get camera language id ***********************************
Byte ps_get_lang_id(void) {return(ps_funcs_system_ptr->subfunc_50());}// see definitions above for returned lang id codes

//**************** get timer ticks since camera power-on in ms **************
DWord ps_get_timer(void)               // note: this function don't reset watchdog
{
  DWord tbuf[2];                       // buffer for QWord value
  ps_funcs_system_ptr->subfunc_bc(&tbuf);
  return(tbuf[0]);                     // return low DWord
}

//**************** wait DELAY_MS miliseconds ********************************
void ps_delay(DWord delay_ms)          // independent on HW speed
{                                      // unlimited delays
  DWord t1,t2;
  t2=ps_get_timer()+delay_ms;          // read current timer value and add required delay
  do {                                 // loop:
    t1=ps_get_timer();                 // read current timer value
    if (t1%(DWord)2000==0)             // if delay exceed 2s
      ps_get_status();                 // prevent camera auto powerdown
    } while (t1<=t2);                  // until it is lower than destination value
}

//**************** get camera system info (firmware version, id string...) **
void ps_get_system_info(PS_SYS_INFO FAR *sysinfo)
{
  Byte FAR *p=ps_funcs_system_ptr->subfunc_00();    // get structure pointer
  FARMEMCPY(sysinfo,p,sizeof(PS_SYS_INFO));         // fill the structure via its pointer
  sysinfo->idstring[sizeof(sysinfo->idstring)-1]=0; // make sure the idstring is zero-terminated
}

//**************** mini key handler - only forwards key state to our program
Word FARPASCAL ps_key_handler(PS_KEY_STATE FAR *keys, DWord FAR *p_keymap)
{
  *p_keymap=(keys->keymap);            // forward keystate to our program
  return(0);
}

//**************** return key state, see definitions above ******************
DWord ps_get_key_state(void)
{
  DWord key=0;                         // 0=anything pressed
  ps_funcs_system_ptr->subfunc_08((PS_KEY_HANDLER FAR *)ps_key_handler,&key);
  return(key);                         // other return key state
}

//**************** wait until given key pressed, also reset watchdog to prevent auto powerdown
void ps_wait_for_key(DWord key_code)
{
  DWord key;
  do {
    key=ps_get_key_state();
  } while ((key&key_code)==0);
}


//**************** subfunctions of API function 5 - firmware ****************
typedef struct
{
  Byte unknown1[0x68];                 // 00-67h
  int  (FARPASCAL *subfunc_68)(int fno, char FAR *ifname, char FAR *ofname); // 68h (extract FW file)
} PS_FUNCS_FIRMWARE;
PS_FUNCS_FIRMWARE FAR *ps_funcs_firmware_ptr;

//**************** extract a file from *.FIR package ************************
int ps_fir_extract(FILE *f, char FAR *ifname, char FAR *ofname) // f is opened *.FIR file
{                                      // ifname is internal file name in package without dot, eg: "diskaimg"
  Word status;                         // ofname is output file name, may include full path, eg "d:\\diska.img"
  int fno=fileno(f);                   // obtain file number of given file
  return(ps_funcs_firmware_ptr->subfunc_68(fno,ifname,ofname)); // retutn 0=OK, -1=invalid ifname
}


//**************** subfunctions of API function 7 - graphics ****************
typedef struct
{
  Word (FARPASCAL *subfunc_00)(Byte manualbuffering, Byte videomode); // 00h (init gfx, vmode)
  void (FARPASCAL *subfunc_04)(void);  // 04h (free video slots)
  void (FARPASCAL *subfunc_08)(Byte layer, Byte sizeindex);// 08h (select bitmap width ?)
  void (FARPASCAL *subfunc_0c)(Byte layer, Byte page);     // 0Ch (select layer and page ?)
  void (FARPASCAL *subfunc_10)(Byte layer, Byte visible);  // 10h (set layer visibility ?)
  void (FARPASCAL *subfunc_14)(Byte layer, Word xo, Word yo); // 14h (set visible offset ?)
  void (FARPASCAL *subfunc_18)(Word FAR *palette);  // 18h (set color palette)
  Word (FARPASCAL *subfunc_1c)(Byte videomode);     // 1Ch (set videomode)
  Word (FARPASCAL *subfunc_20)(Byte method); // 20h (refresh LCD 0=refresh, 2=black screen)
  Byte unknown1[0x08];                 // 24-2Bh
  void (FARPASCAL *subfunc_2c)(Byte enable); // 2Ch (enable autoswitch to TV [0/1])
  Byte (FARPASCAL *subfunc_30)(void);  // 30h (get videomode)
  void (FARPASCAL *subfunc_34)(Word unknown, Word color);  // 34h (clear screen, let unknown=0)
  void (FARPASCAL *subfunc_38)(Word unknown, Word color);  // 38h (fast clear screen, let unknown=0)
  Byte unknown2[0x04];                 // 3C-3Fh
  void (FARPASCAL *subfunc_40)(Word x, Word y, Word w, Word h, Word unknown, Word color); // 40h (filled rectangle, let unknown=0)
  Byte unknown3[0x0C];                 // 44-4Fh
  void (FARPASCAL *subfunc_50)(Word sx, Word sy, Word dx, Word dy, Word w, Word h); // 50h (copy rectangle, y is /2)
  Byte unknown4[0x24];                 // 54-77h
  void (FARPASCAL *subfunc_78)(PS_TEXT_MSG FAR *tm);       // 78h (print message)
} PS_FUNCS_GFX;
PS_FUNCS_GFX FAR *ps_funcs_gfx_ptr;

//**************** init graphics with given videomode and manual buffering flag [0/1]
Word ps_init_gfx(Byte manualbuffering, Byte videomode) {return((ps_funcs_gfx_ptr->subfunc_00(manualbuffering,videomode))&PS_ERR_CODE_MASK);}

//**************** free video slots *****************************************
void ps_free_video_slots(void) {ps_funcs_gfx_ptr->subfunc_04();}

//**************** set videomode (LCD/NTSC/PAL), see definitions above ******
Word ps_set_videomode(Byte videomode) {return((ps_funcs_gfx_ptr->subfunc_1c(videomode))&PS_ERR_CODE_MASK);}

//**************** get videomode, see definitions above *********************
Byte ps_get_videomode(void) {return(ps_funcs_gfx_ptr->subfunc_30());}

//**************** enable automatic switching of display output to TV when plugged in
void ps_enable_auto_switch2tv(Byte enable) {ps_funcs_gfx_ptr->subfunc_2c(enable);}

//**************** refresh LCD from camera screen buffer ********************
Word ps_refresh_lcd(void) {            // need to be called after every gfx function to take effect
  return((ps_funcs_gfx_ptr->subfunc_20(0))&PS_ERR_CODE_MASK);} // 1 refresh takes 16,7ms (60 FPS)

//**************** set LCD palette, which is 32B long array 16x2 ************
void ps_set_pal(Word FAR *palette, Byte refresh)
{
  ps_funcs_gfx_ptr->subfunc_18(palette);
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see changes now
}

//**************** clear camera screen buffer with required color ***********
void ps_cls(Word color, Byte refresh)
{
  ps_funcs_gfx_ptr->subfunc_34(0,color);
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see changes now
}

//**************** clear camera screen buffer with required color fast ******
/*
void ps_fast_cls(Word color, Byte refresh)
{                                      // do the same as subfunc_34
  ps_funcs_gfx_ptr->subfunc_38(0,color);
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see changes now
}*/

//**************** draw filled rectangle by given color [x,y] is top left corner
void ps_draw_frect(Word x, Word y, Word w, Word h, Word color, Byte refresh)
{
  if ((w==0)||(h==0))                  // prevent zero width/height
    return;
  ps_funcs_gfx_ptr->subfunc_40(x,y,w,h,0,color);
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see changes now
}

//**************** draw clipped filled rectangle by given color [x,y] is top left corner
/*
void ps_draw_clipped_frect(int x, int y, int w, int h, Word color, Byte refresh)
{
  if ((x>=PS_SCREEN_XRES)||((x+w)<0)||((y+h)<0)||(y>=PS_SCREEN_YRES))
    return;                            // if totally out of screen return immediately
  if (x<0)
    {w+=x; x=0;}                       // cut off left bound
  if (y<0)
    {h+=y; y=0;}                       // cut off top bound
  if ((x+w)>PS_SCREEN_XRES)
    w=PS_SCREEN_XRES-x;                // cut off right bound
  if ((y+h)>PS_SCREEN_YRES)
    h=PS_SCREEN_YRES-y;                // cut off bottom bound
  ps_draw_frect(x,y,w,h,color,refresh);
}*/

//**************** draw pixel given size and color at [x,y] position (without refresh);
void ps_put_pixel(Word x, Word y, Byte size, Word color) // this is only slow emulation using _frect
{                                      // takes 0,34ms (2937 pix/s) for size 2, 2774 pix/s for size 20, 958 pix/s for size 100
  if ((size==0)||(x>(PS_SCREEN_XRES-size))||(y>(PS_SCREEN_YRES-size))) // prevent zero width/height, clipping
    return;                            // minimum size to look good on LCD is 3 (for less the color gets corrupted due to interpolation)
  ps_funcs_gfx_ptr->subfunc_40(x,y,size,size,0,color);
}

//**************** copy rectangle in buffer from source to destination position
/*
void ps_copy_rect(Word sx, Word sy, Word dx, Word dy, Word w, Word h, Byte refresh)
{
  if ((w==0)||(h==0))                  // prevent zero width/height
    return;                            // we need to divide y coords because
  ps_funcs_gfx_ptr->subfunc_50(sx,sy>>1,dx,dy>>1,w,h>>1); //  it is in range <0,PS_SCREEN_YRES/2> on A70
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see changes now
}*/

//**************** print text message from structure to camera screen buffer
void ps_print_txt_msg_to_buffer(PS_TEXT_MSG FAR *tm) {ps_funcs_gfx_ptr->subfunc_78(tm);}

//**************** display single-line text message on LCD ('\n' not allowed!)
void ps_print(char FAR *str, Word x, Word y, Word txtcolor, Word dmode, Byte refresh)
{                                      // LCD text-mode screen is 32x10 chars
  PS_TEXT_MSG tm;                      // text message structure
  tm.x=x;                              // fill the structure by params
  tm.y=y;
  tm.txtcolor=txtcolor;
  tm.drawmode=dmode;
  tm.bkcolor=0;
  tm.strptr=str;
  ps_print_txt_msg_to_buffer(&tm);     // load structure to camera buffer
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see the text now
}

//**************** display single-line formatted string on LCD (one '\n' allowed at end only!)
void ps_printf(Byte refresh, Word txtcolor, Word dmode, char *formatstring,...)
{                                      // doesn't scroll when bottom of screen reached!
  char msgbuf[80];                     // buffer for formated text messages
  vsprintf(msgbuf,formatstring,...);   // print formatted text to buffer
  if (msgbuf[strlen(msgbuf)-1]=='\n')  // is last char '\n' ?
    msgbuf[strlen(msgbuf)-1]=0;        // remove it
  ps_print(msgbuf,ps_where_x,ps_where_y,txtcolor,dmode,refresh);
  if (formatstring[strlen(formatstring)-1]=='\n')
    {                                  // is last char '\n'
    ps_where_x=0;                      // reset x position
    if (ps_where_y<(PS_SCREEN_YRES-2*PS_FONT_H)) // is not last line?
      ps_where_y+=2*PS_FONT_H;         // move y position one line down
    }
  else                                 // update x position
    ps_where_x+=PS_FONT_W*strlen(msgbuf);
}

//**************** clear one text line by background color (without refresh)
//#define ps_clr_line(line, bkcolor) ps_draw_frect(0,line*2*PS_FONT_H,PS_SCREEN_XRES,2*PS_FONT_H,bkcolor,0)

//**************** draw ^ marker used as cursor for input or selection (obsolete)
/*
void ps_draw_marker(Byte x, Byte y, Word mkcolor, Word bkcolor)
{
  ps_clr_line(y,bkcolor);              // clear line
  if ((x<(PS_SCREEN_XRES/PS_FONT_W))&&(y<(PS_SCREEN_XRES/(2*PS_FONT_H))))
    ps_print("^",x*PS_FONT_W,y*2*PS_FONT_H,mkcolor,0,1); // print '^'
}*/

//**************** draw color cursor used as background for text input/selection (without LCD refresh)
/*
void ps_draw_cursor(Byte x1, Byte x2, Byte curpos, Byte y, Word curcolor, Word bkcolor)
{                                      // check input values, curpos is relative to x1
  if ((y<(PS_SCREEN_YRES/(2*PS_FONT_H)))&&(x1<=x2)&&(x2<(PS_SCREEN_XRES/PS_FONT_W)))
    {                                  // clear background with bkcolor
    ps_draw_frect(x1*PS_FONT_W,y*2*PS_FONT_H,(x2-x1+1)*PS_FONT_W,2*PS_FONT_H,bkcolor,0);
    if ((x1+curpos)<=x2)               // draw color cursor if visible
      ps_draw_frect((x1+curpos)*PS_FONT_W,y*2*PS_FONT_H,PS_FONT_W,2*PS_FONT_H,curcolor,0);
    }
}*/

//**************** set coords position for ps_printf ************************
void ps_gotoxy(Word x, Word y)         // [x,y] uses font char size
{                                      // [0,0] is top left corner
  ps_where_x=x*PS_FONT_W;              // [31,9] is bottom righ corner (on A70)
  ps_where_y=y*2*PS_FONT_H;
}

//**************** subfunctions of API function Ah - menu *******************
typedef struct
{
  Byte unknown1[0x38];                 // 00-37h
  void (FARPASCAL *subfunc_38)(Word x, Word y, Word w, Word h, Word unknown, Word color); // 38h (outlined rectangle, let unknown=0)
} PS_FUNCS_MENU;
PS_FUNCS_MENU FAR *ps_funcs_menu_ptr;

//**************** draw outlined rectangle by given color [x,y] is top left corner
/*
void ps_draw_orect(Word x, Word y, Word w, Word h, Word color, Byte refresh)
{
  if ((w==0)||(h==0))                  // prevent zero width/height
    return;
  ps_funcs_menu_ptr->subfunc_38(x,y,w,h,0,color);
  if (refresh!=0)
    ps_refresh_lcd();                  // refresh LCD to see changes now
}*/

//**************** prepare pointers to API functions in array and structures
void ps_init(void)                     // call this before you use any other PSIOlib function!
{
  int i;
  union REGS regs;                     // generic registers
  struct SREGS segregs;                // segment registers
  for (i=0; i<PS_USED_API_FUNCS; i++)
    {
    regs.h.ah=i;                       // put required funcno to AH
    int86x(0xFF,&regs,&regs,&segregs); // do int FFh with regs structures
    ps_api_functions[i]=MK_FP(segregs.ds,regs.x.dx);
    }                                  // return pointer from DS:DX
  ps_funcs_system_ptr  =(PS_FUNCS_SYSTEM FAR *)ps_api_functions[1]; // init structure pointer by API_1_PTR
  ps_funcs_firmware_ptr=(PS_FUNCS_FIRMWARE FAR *)ps_api_functions[5]; // init structure pointer by API_5_PTR
  ps_funcs_gfx_ptr     =(PS_FUNCS_GFX FAR *)ps_api_functions[7];    // init structure pointer by API_7_PTR
  ps_funcs_menu_ptr    =(PS_FUNCS_MENU FAR *)ps_api_functions[10];  // init structure pointer by API_10_PTR
}

//**************** reboot camera by jumping at FFFF:FFF0 ********************
//#define ps_reboot() {asm {db 0xEA; dd 0xFFFFFFF0}};

//**************** turns the camera speaker on at the specified frequency ***
/*
void ps_sound(Word f)                  // valid range is [61-15625 Hz], 0-turns speaker off
{                                      // the oscillator is based on 64us period
  Word n;                              // every n>0 adds n*64us to base 64us period
  if (f==0)                            // if zero frequency
    outport(0xF05E,0);                 // turn off the Sound System
  if ((f>=PS_NOTE_MINF)&&(f<=PS_NOTE_MAXF)) // check valid frequency range
    {                                  // f=15625/(n+1)
    n=PS_NOTE_MAXF/f-1;                // calculate n [0-255] (base period multiplier)
    outport(0xF05E,(n<<8)|3);          // bits: nnnnnnnnxxxxxxES
    }                                  // E=Enable (1=sound), S=Sound System stand-by (1=active)
} */

//**************** turns the camera speaker off *****************************
//#define ps_nosound() ps_sound(0)

//**************** beep at 1kHz for 0,5s ************************************
//#define ps_beep() {ps_sound(1000); ps_delay(500); ps_sound(0);}

//**************** display camera info on LCD, just a demo of library functions
/*
void ps_display_info(Word textcolor)
{
  struct fatinfo di;                   // FAT info structure
  struct dfree df;                     // disk free info structure
  struct time t;                       // time info structure
  struct date d;                       // date info structure
  PS_SYS_INFO sysinfo;                 // system info structure
  DWord dsize,dfree;
  Byte i;
  ps_get_system_info(&sysinfo);        // get system info
  ps_cls(15,0);                        // clear screen with 15=black
  ps_gotoxy(0,0);                      // set ps_printf position
  ps_printf(0,textcolor,0,"%s\n",sysinfo.idstring); // use transparent bkmode to see color used by cls
  ps_printf(0,textcolor,0,"Firmware version: %d.%d.%d.%d\n",sysinfo.fwver[3],sysinfo.fwver[2],sysinfo.fwver[1],sysinfo.fwver[0]);
  ps_printf(0,textcolor,0,"Camera PID: %04Xh, DOS %d.%d\n",ps_get_pid(),_osmajor,_osminor);
  ps_printf(0,textcolor,0,"Current drive is %c:, Diskinfo:\n",getdisk()+'A');
  for (i=1; i<=4; i++)                 // loop for info about A:..D:
    {
    getfat(i,&di);
    getdfree(i,&df);
    dsize=(DWord)di.fi_nclus*(DWord)di.fi_bysec*(DWord)di.fi_sclus;
    dfree=(DWord)df.df_avail*(DWord)df.df_bsec*(DWord)df.df_sclus;
    ps_printf(0,textcolor,0," %c: %9ld / %9ld B\n",i-1+'A',dsize,dfree);
    }
  gettime(&t);                         // get time & date
  getdate(&d);                         // WARNING: this value is not updated until camera is turned off
  ps_printf(0,textcolor,0,"Date/time: %2d:%02d:%02d, %d.%d.%d\n",t.ti_hour,t.ti_min,t.ti_sec,d.da_day,d.da_mon,d.da_year);
  ps_printf(1,textcolor,0,"Heap Memory Available: %ld B",mem_avail());
}*/

//**************** draw 16 bars with palette colors *************************
/*
void ps_draw_palette(Word textcolor)
{
  int i;
  char msgbuf[30];
  for (i=0; i<16; i++)
    {
    ps_draw_frect(i*(PS_SCREEN_XRES/16),0,PS_SCREEN_XRES/16,PS_SCREEN_YRES,i,0);
    sprintf(msgbuf,"%d",i);
    ps_print(msgbuf,i*(PS_SCREEN_XRES/16),10,textcolor,0,0);
    }
  sprintf(msgbuf,"Current color palette",i);
  ps_print(msgbuf,100,PS_SCREEN_YRES/2,textcolor,0,1);
}*/

//**************** defualt palette with some useful colors ******************
Word ps_default_palette[16]={
  0x3F00, // 0011111100000000 0.  transparent
  0xC003, // 1100000000000011 1.  red
  0xC007, // 1100000000000111 2.  red+ - reserved for text background
  0xC080, // 1100000010000000 3.  green
  0xC088, // 1100000010001000 4.  green+
  0xC010, // 1100000000010000 5.  blue-
  0xC020, // 1100000000100000 6.  blue
  0xC070, // 1100000001110000 7.  blue+
  0xE070, // 1110000001110000 8.  blue++
  0xC022, // 1100000000100010 9.  violet
  0xC044, // 1100000001000100 10. violet+
  0xCA00, // 1100101000000000 11. gray-
  0xD000, // 1101000000000000 12. gray
  0xDC00, // 1101110000000000 13. gray++
  0xFF00, // 1111111100000000 14. white
  0xC000, // 1100000000000000 15. black
};        // OOYYYYYYGBBBGRRR   O-opacity (to black-11), Y-luminance/alpha level (to white-111111)

